unit wavIO;

{ wavIO specializes on the situation where an input wav File is transformed into an output wav file
  with an identical structure but potentially different sample values.
  Reading and writing of sample values is done on a logical block basis (blocksize is configurable and as used in our context is roughly 1000 samples).
  Sample values are always 32 bit integers. With 16 (24) bit files the 16 (24) bit make up for the 16 (24) MSBs amd the remaining bit are set to 0.
  Reading and writing is buffered internally - this is just an internal detail not to be confused with the
  provided reading and writing functions which work on a logical sample block basis.
}

interface

type PLongInt = ^LongInt;

function openWavIO(inWav: string; outWav: string; blocksize: LongInt): boolean;
                                                            // False if operation not succesful.
                                                            // Reads the header chunks of inWav and memorizes the
                                                            // information from the fmt chunk.
                                                            // Creates outWav and fills its header chunks with an
                                                            // identical copy of that of inWav.
                                                            // Allocates the block for reading inWav and writing outWav so that the block is made up
                                                            // of blocksize samples.
function channelCount: integer;                             // number of channels(1: mono, 2: stereo)
                                                            // In the case of stereo the left channel is channel 0 and te right channel is channel 1.
                                                            // In the case of mono the signal is in channel 0 and there is no channel 1.
function sampleRate: LongInt;                               // sample rate in Hz
function bitsPerSample: integer;                            // bits per sample
function readNextSampleBlock(var sampleBlockCh0: array of LongInt; var sampleBlockCh1: array of LongInt; var readsamplesCount: integer): boolean;
                                                            // False if not possible (usually because all the samples have been read).
                                                            // sampleBlockCh0: next block of samples from input wav file for channel 0.
                                                            //                 first sample is in element with index 0.
                                                            // sampleBlockCh1: same for channel 1 if this channel is available.
                                                            // The first call after openWavIO gives the first sample block.
                                                            // readsamplesCount: number of samples read (counting only 1 channel).
function writeNextSampleBlock(var sampleBlockCh0: array of LongInt; var sampleBlockCh1: array of LongInt; samplesCountToBeWritten: integer): boolean;
                                                            // False if not possible.
                                                            // Writes sampleBlockCh0 as the next channel 0 sample block to the output file.
                                                            // First sample written is in element with index 0.
                                                            // Last sample written is in element with index samplesCountToBeWritten - 1.
                                                            // samplesCountToBeWritten should be the readsamplesCount value from the
                                                            // corresponding readNextSampleBlock function call.
                                                            // Same for sampleBlockCh1 in case there is a channel 1. If there is none the values
                                                            // given as sampleCh1 don't matter (sampleBlockCh0 is the natural choice).
function closeWavIO: boolean;                               // False if not possible.
                                                            // Finalizes the wavIO operations.



implementation
uses
  SysUtils;

type PByte = ^Byte;
     PWord = ^Word;
     PSmallInt = ^SmallInt;

var infile, outfile : File;
    filesOpened: boolean;
    inBuf, outBuf: PByte;
    bufsize: integer; // number of byte that make up for inbuf and outbuf
    nrOfByteInOneBlockInBuf: integer;  // number of byte that make up a complete block in inBuf and outBuf.
    bufferAllocated: boolean;
    nrOfFullBlocksReadIntoInBuf: integer;  // with the last BlockRead: how many full blocks managed it into inBuf ?
    nrOfByteOfTheLastIncompleteBlock: integer; // when EOF was encountered with the last BlockRead: the number of byte that didn't make it
                                               // into the last full block
    nrOfBlockInBufNotYetFetched: integer; // the next number of the block in inBuf that was not fetched yet by readNextSampleBlock.
                                          // 0: no data in inBuf;
    nrOfByteFilledInOutBuf: LongInt;      // how many byte have managed it into outBuf that are not written to the outputfile yet?
    chCount: integer;
    samplingrate: LongInt;
    bitdepth : integer;
 
function openWavIO(inWav: string; outWav: string; blocksize: LongInt): boolean; // only standard chunks assumed so far
var
 buf : array[1..44] of char;
 pbuf, pfmt, p: PChar;
 readbyteCount : LongInt;
 nrOfBlocksInBuf: integer;
 begin
 openWavIO := False;
 filesOpened := False;
 bufferAllocated := false;
 {I-}
 AssignFile(infile, inWav);
 if IOResult <> 0 then
    exit;
 Reset(infile, 1);
 if IOResult <> 0 then
    exit;
 AssignFile(outfile, outwav);
 if IOResult <> 0 then
    exit;
 Rewrite(outfile, 1);
 if IOResult <> 0 then
    exit;
 filesOpened := True;
 BlockRead(infile, buf, 44, readbyteCount);
 if (IOResult <> 0) or (readbyteCount <> 44) then
    exit;
 BlockWrite(outfile, buf, 44);
 if IOResult <> 0 then
    exit;
 {I+}
 chCount := 0;
 samplingrate := 0;
 bitdepth := 0;
 if copy(buf, 1, 4) <> 'RIFF' then
    begin
    writeln('No RIFF file.');
    closeWavIO;
    exit;
    end;
 if copy(buf, 9, 4) <> 'WAVE' then
    begin
    writeln('No WAVE file.');
    closeWavIO;
    exit;
    end;
 if copy(buf, 13, 4) <> 'fmt ' then
    begin
    writeln('No instant fmt chunk in wav header.');
    closeWavIO;
    exit;
    end;
 if copy(buf, 37, 4) <> 'data' then
    begin
    writeln('data chunk not immediately following fmt chunk in wav header.');
    closeWavIO;
    exit;
    end;
 pbuf := @buf[1];
 pfmt := pbuf;
 Inc(pfmt, 12);   // pfmt points to fmt chunk
 p := pfmt;
 Inc(p, 8);       // points to data format in fmt chunk. content should be 1 (PCM format)
 if PWord(p)^ <> 1 then
    begin
    writeln('data chunk not immediately following fmt chunk in wav header.');
    closeWavIO;
    exit;
    end;
 p := pfmt;
 Inc(p, 10);      // points to number of channels in fmt chunk. content should be 1 (PCM format)
 chCount := PWord(p)^;
 p := pfmt;
 Inc(p, 12);      // points to sample rate in fmt chunk. content should be 1 (PCM format)
 samplingrate := PLongInt(p)^;
 if (samplingrate < 32000) or (samplingrate > 48000) then
    begin
    writeln('Sample rate not supported.');
    writeln('Should be between 32000 and 48000 Hz.');
    closeWavIO;
    exit;
    end;
 p := pfmt;
 Inc(p, 22);      // points to bits per sample in fmt chunk. content should be 1 (PCM format)
 bitdepth := PWord(p)^;
 if (bitdepth <> 16) and (bitdepth <> 24) then
    begin
    writeln('Input file has a bitrate of ' + IntToStr(bitdepth) + ' bit per sample.');
    writeln('This is not supported.');
    closeWavIO;
    exit;
    end;
 if blocksize < 1100 then
    nrOfBlocksInBuf := 10
 else
    if blocksize < 2200 then
       nrOfBlocksInBuf := 5
    else
       nrOfBlocksInBuf := 1;
 nrOfByteInOneBlockInBuf := (bitdepth DIV 8) * chCount * blocksize;
 bufsize := nrOfByteInOneBlockInBuf * nrOfBlocksInBuf;
 GetMem(inBuf, bufsize);
 GetMem(outBuf, bufsize);
 bufferAllocated := false;
 nrOfFullBlocksReadIntoInBuf := 0;
 nrOfByteOfTheLastIncompleteBlock := 0;
 nrOfBlockInBufNotYetFetched := 0;
 nrOfByteFilledInOutBuf := 0;
 openWavIO := True;
 end;

function channelCount: integer;
 begin
   channelCount := chCount;
 end;

function sampleRate: LongInt;
 begin
   sampleRate := samplingrate;
 end;

function bitsPerSample: integer;
 begin
   bitsPerSample := bitdepth;
 end;

function readNextSampleBlock(var sampleBlockCh0: array of LongInt; var sampleBlockCh1: array of LongInt; var readsamplesCount: integer): boolean;
 var
 readbyteCount: LongInt;
 pB, pEndB: PByte;
 pInt16, pEndInt16: PSmallInt;
 LoByte: Byte;
 i: integer;
 begin
   readNextSampleBlock := False;
   if nrOfBlockInBufNotYetFetched = 0 then
      begin
      if nrOfByteOfTheLastIncompleteBlock > 0 then
         exit; // no more data
      {I-}
      BlockRead(infile, inbuf^, bufsize, readbyteCount);
      if IOResult <> 0 then
         exit;
      {I+}
      if readbyteCount = 0 then
         exit; // no more data
      nrOfFullBlocksReadIntoInBuf := readbyteCount DIV nrOfByteInOneBlockInBuf;
      nrOfByteOfTheLastIncompleteBlock := readbyteCount - nrOfFullBlocksReadIntoInBuf * nrOfByteInOneBlockInBuf;
      nrOfBlockInBufNotYetFetched := 1;
      end;
   pB := inBuf;
   Inc(pB, (nrOfBlockInBufNotYetFetched - 1) * nrOfByteInOneBlockInBuf);
   pInt16 := PSmallInt(pB); // points to first sample in Block
   if nrOfBlockInBufNotYetFetched <= nrOfFullBlocksReadIntoInBuf then
      begin
      Inc(pB, nrOfByteInOneBlockInBuf);  // pB points to the spot after last sample in Block
      readsamplesCount := nrOfByteInOneBlockInBuf DIV ((bitdepth DIV 8) * chCount);
      end
   else
      begin
      Inc(pB, nrOfByteOfTheLastIncompleteBlock);  // pB points to the spot after last sample in Block
      nrOfBlockInBufNotYetFetched := -1; // force a try for a new Blockread with next call which will lead to a return False
      readsamplesCount := nrOfByteOfTheLastIncompleteBlock DIV ((bitdepth DIV 8) * chCount);
      end;
   i := 0;
   if bitdepth = 16 then
      begin
      pEndInt16 := PSmallInt(pB);
      if chCount = 1 then
         repeat
            sampleBlockCh0[i] := 65536 * pInt16^;
            Inc(pInt16);
            i := i +1;
         until pInt16 = pEndInt16
      else // stereo:
         repeat
            sampleBlockCh0[i] := 65536 * pInt16^;
            Inc(pInt16);
            sampleBlockCh1[i] := 65536 * pInt16^;
            Inc(pInt16);
            i := i +1;
         until pInt16 = pEndInt16;
      end
   else  // 24 bit
      begin
      pEndB := PByte(pB);
      pB := PByte(pInt16);
      if chCount = 1 then
         repeat
            loByte := pB^;
            Inc(pB);
            pInt16 := PSmallInt(pB);
            Inc(pB, 2);
            sampleBlockCh0[i] := 65536 * pInt16^ + 256 * loByte;
            i := i +1;
         until pB = pEndB
      else // stereo:
         repeat
            loByte := pB^;
            Inc(pB);
            pInt16 := PSmallInt(pB);
            Inc(pB, 2);
            sampleBlockCh0[i] := 65536 * pInt16^ + 256 * loByte;
            loByte := pB^;
            Inc(pB);
            pInt16 := PSmallInt(pB);
            Inc(pB, 2);
            sampleBlockCh1[i] := 65536 * pInt16^ + 256 * loByte;
            i := i +1;
         until pB = pEndB
      end;
   nrOfBlockInBufNotYetFetched := nrOfBlockInBufNotYetFetched + 1;
   if (nrOfBlockInBufNotYetFetched > nrOfFullBlocksReadIntoInBuf) and (nrOfByteOfTheLastIncompleteBlock = 0) then
      nrOfBlockInBufNotYetFetched := 0;
   readNextSampleBlock := True;
 end;

procedure Int24ToInt16andLoByte(int24: LongInt; var hiInt16: SmallInt; var loByte: Byte);
// done in such a way that: 24bit = 256 * hiInt + loByte such that loByte >= 0;
var
 loByteTest: SmallInt;
begin
 hiInt16 := int24 DIV 256;
 loByteTest := int24 - 256 * hiInt16;
 if loByteTest >= 0 then
    loByte := Byte(loByteTest)
 else
    begin
    hiInt16 := hiInt16 - 1;
    loByte := Byte(loByteTest + 256);
    end;
end;

function writeNextSampleBlock(var sampleBlockCh0: array of LongInt; var sampleBlockCh1: array of LongInt; samplesCountToBeWritten: integer): boolean;
var
 nrOfByteForOutBuf, sample: LongInt;
 pB: PByte;
 pInt16: PSmallInt;
 i: integer;
 sampleLoByte: Byte;
 sampleHiInt: SmallInt;
 begin
   writeNextSampleBlock := False;
   nrOfByteForOutBuf := (bitdepth DIV 8) * chCount * samplesCountToBeWritten;
   if nrOfByteFilledInOutBuf + nrOfByteForOutBuf > bufsize then // not sufficient space in outBut: flush outBuf
      begin
      {I-}
      BlockWrite(outfile, outBuf^, nrOfByteFilledInOutBuf);
      if IOResult <> 0 then
         exit;
      {I+}
      nrOfByteFilledInOutBuf := 0;
      end;
   pB := outBuf;
   Inc(pB, nrOfByteFilledInOutBuf); // points to first free spot in outBuf
   if bitdepth = 16 then
      begin
      pInt16 := PSmallInt(pB);
      if chCount = 1 then
         for i:= 0 to samplesCountToBeWritten -1 do
             begin
             pInt16^ := sampleBlockCh0[i] DIV 65536;
             Inc(pInt16);
             end
      else // stereo:
         for i:= 0 to samplesCountToBeWritten -1 do
             begin
             pInt16^ := sampleBlockCh0[i] DIV 65536;
             Inc(pInt16);
             pInt16^ := sampleBlockCh1[i] DIV 65536;
             Inc(pInt16);
             end
      end
   else // 24 bit
      begin
      if chCount = 1 then
         for i:= 0 to samplesCountToBeWritten -1 do
             begin
             sample := sampleBlockCh0[i] DIV 256; // 24 bit value
             Int24ToInt16andLoByte(sample, sampleHiInt, sampleLoByte);
             pB^ := sampleLoByte;
             Inc(pB);
             pInt16 := PSmallInt(pB);
             pInt16^ := sampleHiInt;
             Inc(pB, 2);
             end
      else // stereo:
         for i:= 0 to samplesCountToBeWritten -1 do
             begin
             sample := sampleBlockCh0[i] DIV 256; // 24 bit value
             Int24ToInt16andLoByte(sample, sampleHiInt, sampleLoByte);
             pB^ := sampleLoByte;
             Inc(pB);
             pInt16 := PSmallInt(pB);
             pInt16^ := sampleHiInt;
             Inc(pB, 2);
             sample := sampleBlockCh1[i] DIV 256; // 24 bit value
             Int24ToInt16andLoByte(sample, sampleHiInt, sampleLoByte);
             pB^ := sampleLoByte;
             Inc(pB);
             pInt16 := PSmallInt(pB);
             pInt16^ := sampleHiInt;
             Inc(pB, 2);
             end;
      end;
   nrOfByteFilledInOutBuf := nrOfByteFilledInOutBuf + nrOfByteForOutBuf;
   writeNextSampleBlock := True;
 end;

function closeWavIO: boolean;
 begin
   closeWavIO := False;
   if filesOpened then
      begin
      if nrOfByteFilledInOutBuf > 0 then
         begin
         {I-}
         BlockWrite(outfile, outBuf^, nrOfByteFilledInOutBuf);
         if IOResult <> 0 then
            exit;
         {I+}
         end;
      {I-}
      closeFile(infile);
      if IOResult <> 0 then
         exit;
      CloseFile(outfile);
      if IOResult <> 0 then
         exit;
      {I+}
      filesOpened := False;
   end;
   if bufferAllocated then
      begin
      ReAllocMem(inBuf, 0);
      ReAllocMem(outBuf, 0);
      bufferAllocated := False;
      end;
   closeWavIO := True;
 end;

end.
